package org.apache.dolphinscheduler.plugin.task.datax;

package ..deTemp.1736068417556;

import com.alibaba.druid.sql.ast.SQLStatement;
import com.alibaba.druid.sql.ast.expr.SQLIdentifierExpr;
import com.alibaba.druid.sql.ast.expr.SQLPropertyExpr;
import com.alibaba.druid.sql.ast.statement.SQLSelect;
import com.alibaba.druid.sql.ast.statement.SQLSelectItem;
import com.alibaba.druid.sql.ast.statement.SQLSelectQueryBlock;
import com.alibaba.druid.sql.ast.statement.SQLSelectStatement;
import com.alibaba.druid.sql.ast.statement.SQLUnionQuery;
import com.alibaba.druid.sql.parser.SQLStatementParser;
import com.fasterxml.jackson.databind.node.ArrayNode;
import com.fasterxml.jackson.databind.node.ObjectNode;
import java.io.File;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.OpenOption;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.nio.file.attribute.FileAttribute;
import java.nio.file.attribute.PosixFilePermission;
import java.nio.file.attribute.PosixFilePermissions;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.SystemUtils;
import org.apache.dolphinscheduler.plugin.datasource.api.plugin.DataSourceClientProvider;
import org.apache.dolphinscheduler.plugin.datasource.api.utils.DataSourceUtils;
import org.apache.dolphinscheduler.plugin.datasource.api.utils.PasswordUtils;
import org.apache.dolphinscheduler.plugin.task.api.AbstractTaskExecutor;
import org.apache.dolphinscheduler.plugin.task.api.ShellCommandExecutor;
import org.apache.dolphinscheduler.plugin.task.api.TaskExecutionContext;
import org.apache.dolphinscheduler.plugin.task.api.model.Property;
import org.apache.dolphinscheduler.plugin.task.api.model.TaskResponse;
import org.apache.dolphinscheduler.plugin.task.api.parameters.AbstractParameters;
import org.apache.dolphinscheduler.plugin.task.api.parser.ParamUtils;
import org.apache.dolphinscheduler.plugin.task.api.parser.ParameterUtils;
import org.apache.dolphinscheduler.plugin.task.datax.DataxParameters;
import org.apache.dolphinscheduler.plugin.task.datax.DataxTaskExecutionContext;
import org.apache.dolphinscheduler.plugin.task.datax.DataxUtils;
import org.apache.dolphinscheduler.spi.datasource.BaseConnectionParam;
import org.apache.dolphinscheduler.spi.datasource.ConnectionParam;
import org.apache.dolphinscheduler.spi.enums.DbType;
import org.apache.dolphinscheduler.spi.enums.Flag;
import org.apache.dolphinscheduler.spi.utils.JSONUtils;
import org.apache.dolphinscheduler.spi.utils.StringUtils;

public class DataxTask extends AbstractTaskExecutor {
  public static final String JVM_PARAM = " --jvm=\"-Xms%sG -Xmx%sG\" ";
  
  public static final String CUSTOM_PARAM = " -D%s='%s'";
  
  private static final String DATAX_PYTHON = "python2.7";
  
  private static final Pattern PYTHON_PATH_PATTERN = Pattern.compile("/bin/python[\\d.]*$");
  
  private static final String DATAX_PATH = "${DATAX_HOME}/bin/datax.py";
  
  private static final int DATAX_CHANNEL_COUNT = 1;
  
  private DataxParameters dataXParameters;
  
  private ShellCommandExecutor shellCommandExecutor;
  
  private TaskExecutionContext taskExecutionContext;
  
  private DataxTaskExecutionContext dataxTaskExecutionContext;
  
  public DataxTask(TaskExecutionContext taskExecutionContext) {
    super(taskExecutionContext);
    this.taskExecutionContext = taskExecutionContext;
    this.shellCommandExecutor = new ShellCommandExecutor(this::logHandle, taskExecutionContext, this.logger);
  }
  
  public void init() {
    this.logger.info("datax task params {}", this.taskExecutionContext.getTaskParams());
    this.dataXParameters = (DataxParameters)JSONUtils.parseObject(this.taskExecutionContext.getTaskParams(), DataxParameters.class);
    if (!this.dataXParameters.checkParameters())
      throw new RuntimeException("datax task params is not valid"); 
    this.dataxTaskExecutionContext = this.dataXParameters.generateExtendedContext(this.taskExecutionContext.getResourceParametersHelper());
  }
  
  public void handle() throws Exception {
    try {
      Map<String, Property> paramsMap = this.taskExecutionContext.getPrepareParamsMap();
      String jsonFilePath = buildDataxJsonFile(paramsMap);
      String shellCommandFilePath = buildShellCommandFile(jsonFilePath, paramsMap);
      TaskResponse commandExecuteResult = this.shellCommandExecutor.run(shellCommandFilePath);
      setExitStatusCode(commandExecuteResult.getExitStatusCode());
      setAppIds(String.join(",", getApplicationIds()));
      setProcessId(commandExecuteResult.getProcessId());
    } catch (Exception e) {
      this.logger.error("datax task error", e);
      setExitStatusCode(-1);
      throw e;
    } 
  }
  
  public void cancelApplication(boolean cancelApplication) throws Exception {
    this.shellCommandExecutor.cancelApplication();
  }
  
  private String buildDataxJsonFile(Map<String, Property> paramsMap) throws Exception {
    String fileName = String.format("%s/%s_job.json", new Object[] { this.taskExecutionContext
          .getExecutePath(), this.taskExecutionContext
          .getTaskAppId() });
    Path path = (new File(fileName)).toPath();
    if (Files.exists(path, new java.nio.file.LinkOption[0]))
      return fileName; 
    if (this.dataXParameters.getCustomConfig() == Flag.YES.ordinal()) {
      json = this.dataXParameters.getJson().replaceAll("\\r\\n", "\n");
    } else {
      ObjectNode job = JSONUtils.createObjectNode();
      job.putArray("content").addAll((Collection)buildDataxJobContentJson());
      job.set("setting", buildDataxJobSettingJson());
      ObjectNode root = JSONUtils.createObjectNode();
      root.set("job", job);
      root.set("core", buildDataxCoreJson());
      json = root.toString();
    } 
    String json = ParameterUtils.convertParameterPlaceholders(json, ParamUtils.convert(paramsMap));
    this.logger.debug("datax job json : {}", json);
    FileUtils.writeStringToFile(new File(fileName), json, StandardCharsets.UTF_8);
    return fileName;
  }
  
  private List<ObjectNode> buildDataxJobContentJson() {
    BaseConnectionParam dataSourceCfg = (BaseConnectionParam)DataSourceUtils.buildConnectionParams(this.dataxTaskExecutionContext
        .getSourcetype(), this.dataxTaskExecutionContext
        .getSourceConnectionParams());
    BaseConnectionParam dataTargetCfg = (BaseConnectionParam)DataSourceUtils.buildConnectionParams(this.dataxTaskExecutionContext
        .getTargetType(), this.dataxTaskExecutionContext
        .getTargetConnectionParams());
    List<ObjectNode> readerConnArr = new ArrayList<>();
    ObjectNode readerConn = JSONUtils.createObjectNode();
    ArrayNode sqlArr = readerConn.putArray("querySql");
    for (String sql : new String[] { this.dataXParameters.getSql() })
      sqlArr.add(sql); 
    ArrayNode urlArr = readerConn.putArray("jdbcUrl");
    urlArr.add(DataSourceUtils.getJdbcUrl(DbType.valueOf(this.dataXParameters.getDsType()), (ConnectionParam)dataSourceCfg));
    readerConnArr.add(readerConn);
    ObjectNode readerParam = JSONUtils.createObjectNode();
    readerParam.put("username", dataSourceCfg.getUser());
    readerParam.put("password", PasswordUtils.decodePassword(dataSourceCfg.getPassword()));
    readerParam.putArray("connection").addAll((Collection)readerConnArr);
    ObjectNode reader = JSONUtils.createObjectNode();
    reader.put("name", DataxUtils.getReaderPluginName(this.dataxTaskExecutionContext.getSourcetype()));
    reader.set("parameter", readerParam);
    List<ObjectNode> writerConnArr = new ArrayList<>();
    ObjectNode writerConn = JSONUtils.createObjectNode();
    ArrayNode tableArr = writerConn.putArray("table");
    tableArr.add(this.dataXParameters.getTargetTable());
    writerConn.put("jdbcUrl", DataSourceUtils.getJdbcUrl(DbType.valueOf(this.dataXParameters.getDtType()), (ConnectionParam)dataTargetCfg));
    writerConnArr.add(writerConn);
    ObjectNode writerParam = JSONUtils.createObjectNode();
    writerParam.put("username", dataTargetCfg.getUser());
    writerParam.put("password", PasswordUtils.decodePassword(dataTargetCfg.getPassword()));
    String[] columns = parsingSqlColumnNames(this.dataxTaskExecutionContext.getSourcetype(), this.dataxTaskExecutionContext
        .getTargetType(), dataSourceCfg, this.dataXParameters
        .getSql());
    ArrayNode columnArr = writerParam.putArray("column");
    for (String column : columns)
      columnArr.add(column); 
    writerParam.putArray("connection").addAll((Collection)writerConnArr);
    if (CollectionUtils.isNotEmpty(this.dataXParameters.getPreStatements())) {
      ArrayNode preSqlArr = writerParam.putArray("preSql");
      for (String preSql : this.dataXParameters.getPreStatements())
        preSqlArr.add(preSql); 
    } 
    if (CollectionUtils.isNotEmpty(this.dataXParameters.getPostStatements())) {
      ArrayNode postSqlArr = writerParam.putArray("postSql");
      for (String postSql : this.dataXParameters.getPostStatements())
        postSqlArr.add(postSql); 
    } 
    ObjectNode writer = JSONUtils.createObjectNode();
    writer.put("name", DataxUtils.getWriterPluginName(this.dataxTaskExecutionContext.getTargetType()));
    writer.set("parameter", writerParam);
    List<ObjectNode> contentList = new ArrayList<>();
    ObjectNode content = JSONUtils.createObjectNode();
    content.set("reader", reader);
    content.set("writer", writer);
    contentList.add(content);
    return contentList;
  }
  
  private ObjectNode buildDataxJobSettingJson() {
    ObjectNode speed = JSONUtils.createObjectNode();
    speed.put("channel", 1);
    if (this.dataXParameters.getJobSpeedByte() > 0)
      speed.put("byte", this.dataXParameters.getJobSpeedByte()); 
    if (this.dataXParameters.getJobSpeedRecord() > 0)
      speed.put("record", this.dataXParameters.getJobSpeedRecord()); 
    ObjectNode errorLimit = JSONUtils.createObjectNode();
    errorLimit.put("record", 0);
    errorLimit.put("percentage", 0);
    ObjectNode setting = JSONUtils.createObjectNode();
    setting.set("speed", speed);
    setting.set("errorLimit", errorLimit);
    return setting;
  }
  
  private ObjectNode buildDataxCoreJson() {
    ObjectNode speed = JSONUtils.createObjectNode();
    speed.put("channel", 1);
    if (this.dataXParameters.getJobSpeedByte() > 0)
      speed.put("byte", this.dataXParameters.getJobSpeedByte()); 
    if (this.dataXParameters.getJobSpeedRecord() > 0)
      speed.put("record", this.dataXParameters.getJobSpeedRecord()); 
    ObjectNode channel = JSONUtils.createObjectNode();
    channel.set("speed", speed);
    ObjectNode transport = JSONUtils.createObjectNode();
    transport.set("channel", channel);
    ObjectNode core = JSONUtils.createObjectNode();
    core.set("transport", transport);
    return core;
  }
  
  private String buildShellCommandFile(String jobConfigFilePath, Map<String, Property> paramsMap) throws Exception {
    String fileName = String.format("%s/%s_node.%s", new Object[] { this.taskExecutionContext
          .getExecutePath(), this.taskExecutionContext
          .getTaskAppId(), SystemUtils.IS_OS_WINDOWS ? "bat" : "sh" });
    Path path = (new File(fileName)).toPath();
    if (Files.exists(path, new java.nio.file.LinkOption[0]))
      return fileName; 
    StringBuilder sbr = new StringBuilder();
    sbr.append(getPythonCommand());
    sbr.append(" ");
    sbr.append("${DATAX_HOME}/bin/datax.py");
    sbr.append(" ");
    sbr.append(loadJvmEnv(this.dataXParameters));
    sbr.append(addCustomParameters(paramsMap));
    sbr.append(" ");
    sbr.append(jobConfigFilePath);
    String dataxCommand = ParameterUtils.convertParameterPlaceholders(sbr.toString(), ParamUtils.convert(paramsMap));
    this.logger.info("raw script : {}", dataxCommand);
    Set<PosixFilePermission> perms = PosixFilePermissions.fromString("rwxr-xr-x");
    FileAttribute<Set<PosixFilePermission>> attr = PosixFilePermissions.asFileAttribute(perms);
    if (SystemUtils.IS_OS_WINDOWS) {
      Files.createFile(path, (FileAttribute<?>[])new FileAttribute[0]);
    } else {
      Files.createFile(path, (FileAttribute<?>[])new FileAttribute[] { attr });
    } 
    Files.write(path, dataxCommand.getBytes(), new OpenOption[] { StandardOpenOption.APPEND });
    return fileName;
  }
  
  private StringBuilder addCustomParameters(Map<String, Property> paramsMap) {
    StringBuilder customParameters = new StringBuilder("-p\"");
    for (Map.Entry<String, Property> entry : paramsMap.entrySet()) {
      customParameters.append(String.format(" -D%s='%s'", new Object[] { entry.getKey(), ((Property)entry.getValue()).getValue() }));
    } 
    customParameters.append("\"");
    return customParameters;
  }
  
  public String getPythonCommand() {
    String pythonHome = System.getenv("PYTHON_HOME");
    return getPythonCommand(pythonHome);
  }
  
  public String getPythonCommand(String pythonHome) {
    if (StringUtils.isEmpty(pythonHome))
      return "python2.7"; 
    String pythonBinPath = "/bin/python2.7";
    Matcher matcher = PYTHON_PATH_PATTERN.matcher(pythonHome);
    if (matcher.find())
      return matcher.replaceAll(pythonBinPath); 
    return Paths.get(pythonHome, new String[] { pythonBinPath }).toString();
  }
  
  public String loadJvmEnv(DataxParameters dataXParameters) {
    int xms = Math.max(dataXParameters.getXms(), 1);
    int xmx = Math.max(dataXParameters.getXmx(), 1);
    return String.format(" --jvm=\"-Xms%sG -Xmx%sG\" ", new Object[] { Integer.valueOf(xms), Integer.valueOf(xmx) });
  }
  
  private String[] parsingSqlColumnNames(DbType sourceType, DbType targetType, BaseConnectionParam dataSourceCfg, String sql) {
    String[] columnNames = tryGrammaticalAnalysisSqlColumnNames(sourceType, sql);
    if (columnNames == null || columnNames.length == 0) {
      this.logger.info("try to execute sql analysis query column name");
      columnNames = tryExecuteSqlResolveColumnNames(sourceType, dataSourceCfg, sql);
    } 
    notNull(columnNames, String.format("parsing sql columns failed : %s", new Object[] { sql }));
    return DataxUtils.convertKeywordsColumns(targetType, columnNames);
  }
  
  private String[] tryGrammaticalAnalysisSqlColumnNames(DbType dbType, String sql) {
    String[] columnNames;
    try {
      SQLStatementParser parser = DataxUtils.getSqlStatementParser(dbType, sql);
      if (parser == null) {
        this.logger.warn("database driver [{}] is not support grammatical analysis sql", dbType);
        return new String[0];
      } 
      SQLStatement sqlStatement = parser.parseStatement();
      SQLSelectStatement sqlSelectStatement = (SQLSelectStatement)sqlStatement;
      SQLSelect sqlSelect = sqlSelectStatement.getSelect();
      List<SQLSelectItem> selectItemList = null;
      if (sqlSelect.getQuery() instanceof SQLSelectQueryBlock) {
        SQLSelectQueryBlock block = (SQLSelectQueryBlock)sqlSelect.getQuery();
        selectItemList = block.getSelectList();
      } else if (sqlSelect.getQuery() instanceof SQLUnionQuery) {
        SQLUnionQuery unionQuery = (SQLUnionQuery)sqlSelect.getQuery();
        SQLSelectQueryBlock block = (SQLSelectQueryBlock)unionQuery.getRight();
        selectItemList = block.getSelectList();
      } 
      notNull(selectItemList, 
          String.format("select query type [%s] is not support", new Object[] { sqlSelect.getQuery().toString() }));
      columnNames = new String[selectItemList.size()];
      for (int i = 0; i < selectItemList.size(); i++) {
        SQLSelectItem item = selectItemList.get(i);
        String columnName = null;
        if (item.getAlias() != null) {
          columnName = item.getAlias();
        } else if (item.getExpr() != null) {
          if (item.getExpr() instanceof SQLPropertyExpr) {
            SQLPropertyExpr expr = (SQLPropertyExpr)item.getExpr();
            columnName = expr.getName();
          } else if (item.getExpr() instanceof SQLIdentifierExpr) {
            SQLIdentifierExpr expr = (SQLIdentifierExpr)item.getExpr();
            columnName = expr.getName();
          } 
        } else {
          throw new RuntimeException(
              String.format("grammatical analysis sql column [ %s ] failed", new Object[] { item.toString() }));
        } 
        if (columnName == null)
          throw new RuntimeException(
              String.format("grammatical analysis sql column [ %s ] failed", new Object[] { item.toString() })); 
        columnNames[i] = columnName;
      } 
    } catch (Exception e) {
      this.logger.warn(e.getMessage(), e);
      return new String[0];
    } 
    return columnNames;
  }
  
  public String[] tryExecuteSqlResolveColumnNames(DbType sourceType, BaseConnectionParam baseDataSource, String sql) {
    String[] columnNames;
    sql = String.format("SELECT t.* FROM ( %s ) t WHERE 0 = 1", new Object[] { sql });
    sql = sql.replace(";", "");
    try(Connection connection = DataSourceClientProvider.getInstance().getConnection(sourceType, (ConnectionParam)baseDataSource); 
        PreparedStatement stmt = connection.prepareStatement(sql); 
        ResultSet resultSet = stmt.executeQuery()) {
      ResultSetMetaData md = resultSet.getMetaData();
      int num = md.getColumnCount();
      columnNames = new String[num];
      for (int i = 1; i <= num; i++)
        columnNames[i - 1] = md.getColumnName(i); 
    } catch (SQLException|java.util.concurrent.ExecutionException e) {
      this.logger.error(e.getMessage(), e);
      return null;
    } 
    return columnNames;
  }
  
  public AbstractParameters getParameters() {
    return (AbstractParameters)this.dataXParameters;
  }
  
  private void notNull(Object obj, String message) {
    if (obj == null)
      throw new RuntimeException(message); 
  }
}
